

import os
import re
import shutil
from twisted.internet.defer import inlineCallbacks
from twisted.internet.threads import deferToThread

from ebs.linuxnode.exim.mixin import LocalEximMixin
from ebs.linuxnode.exim.local import ImportSpec
from ebs.linuxnode.core.config import ElementSpec, ItemSpec
from ebs.linuxnode.core.background import BackgroundSpec

from .basic import BackgroundSequenceMixin


class BackgroundSequenceImporterMixin(BackgroundSequenceMixin, LocalEximMixin):
    _regex_duration = re.compile(r".(?P<numeric>\d+)(?P<unit>[smh]).")
    _regex_structured = re.compile(r"^(\d+_)?(?P<target>structured:\S*?)(?P<duration>\.(?P<numeric>\d+)(?P<unit>[smh]))?$")  # noqa
    _multipliers = {'s': 1, 'm': 60, 'h': 3600}
    _force_duration_ext = ('.png', '.jpg', '.bmp', '.gif', '.jpeg')

    def __init__(self, *args, **kwargs):
        super(BackgroundSequenceImporterMixin, self).__init__(*args, **kwargs)
        self._allowed_structured_backgrounds = []

    def register_structured_background(self, background):
        self._allowed_structured_backgrounds.append(background)

    @property
    def _bg_local_path(self):
        return os.path.join(self.cache_dir, 'bg')

    @inlineCallbacks
    def _import_backgrounds(self, source_path):
        self.log.info("Importing Backgrounds from '{}' to '{}'"
                      "".format(source_path, self._bg_local_path))
        if not os.path.exists(self._bg_local_path):
            os.makedirs(self._bg_local_path, exist_ok=True)
        yield self.exim.clear_directory(self._bg_local_path)
        _targets = []
        for filename in sorted(os.listdir(source_path)):
            source_filepath = os.path.join(source_path, filename)
            dest_filepath = os.path.join(self._bg_local_path, filename)
            if not os.path.isfile(source_filepath):
                continue

            self.log.info("Copying {} from {} to {}"
                          "".format(filename, source_path, self._bg_local_path))
            yield deferToThread(shutil.copy, source_filepath, dest_filepath)

            _duration = None

            structured = self._regex_structured.match(filename)
            if structured:
                target = structured.groupdict()['target']
                if filename.split(':')[1] in self._allowed_structured_backgrounds:
                    dest_filepath = target
                else:
                    continue
                if structured.groupdict()['duration']:
                    unit = structured.groupdict()['unit']
                    multiplier = self._multipliers[unit]
                    _duration = structured.groupdict()['numeric'] * multiplier
                else:
                    _duration = 30

            if not _duration:
                _duration = self._regex_duration.search(filename, re.IGNORECASE)
                if _duration:
                    unit = _duration.groupdict()['unit']
                    multiplier = self._multipliers[unit]
                    _duration = _duration.groupdict()['numeric'] * multiplier

            if not _duration:
                ext = os.path.splitext(filename)[1]
                if ext and ext in self._force_duration_ext:
                    _duration = 10

            if _duration:
                _target = BackgroundSpec(dest_filepath, duration=_duration)
            else:
                _target = dest_filepath
            _targets.append(_target)
        self.background_sequence_set(_targets)

    def install(self):
        super(BackgroundSequenceImporterMixin, self).install()
        _elements = {
            'exim_local_background': ElementSpec('exim', 'local_background', ItemSpec(bool, fallback=False)),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)
        target_path = self._bg_local_path
        if self.config.exim_local_background:
            spec = ImportSpec(target_path, source='[id]?',
                              writer=self._import_backgrounds,
                              contexts=['startup'])
            self.exim.register_import('bg', spec)





from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import Text
from sqlalchemy.ext.declarative import declarative_base

from ebs.linuxnode.core.background import BackgroundSpec

Base = declarative_base()
metadata = Base.metadata


class BackgroundModel(Base):
    __tablename__ = 'bg'

    id = Column(Integer, primary_key=True)
    seq = Column(Integer, unique=True, index=True)
    target = Column(Text)
    bgcolor = Column(Text, nullable=True)
    duration = Column(Integer, nullable=True)

    def __init__(self, seq: int, spec: BackgroundSpec):
        self.seq = seq
        self.target = spec.target
        self.bgcolor = spec.bgcolor
        self.duration = spec.duration

    def native(self):
        return BackgroundSpec(target=self.target, bgcolor=self.bgcolor,
                              callback=None, duration=self.duration)

    def __repr__(self):
        return "{0:3} {1} {2} [{3}]".format(
            self.seq, self.target, self.bgcolor, self.duration or ''
        )





from itertools import cycle

from ebs.linuxnode.core.background import BackgroundCoreMixin
from ebs.linuxnode.core.background import BackgroundSpec

from .persistence import BackgroundSequencePersistenceManager


class BackgroundSequenceMixin(BackgroundCoreMixin):
    def __init__(self, *args, **kwargs):
        super(BackgroundSequenceMixin, self).__init__(*args, **kwargs)
        self._bg_sequence = cycle([])
        self._bg_sequence_persistence = BackgroundSequencePersistenceManager(self)
        self._bg_sequence_active = False

    @property
    def bg_sequence(self):
        return self._bg_sequence

    @bg_sequence.setter
    def bg_sequence(self, value):
        self._bg_sequence = cycle(value)
        if not self._bg_sequence_active:
            self.bg_step()
        else:
            ntargets = []
            for item in value:
                if isinstance(item, BackgroundSpec):
                    ntargets.append(item.target)
                else:
                    ntargets.append(item)
            if self._bg_current in ntargets:
                curr = None
                while curr != self._bg_current:
                    curr = next(self._bg_sequence)
                    if isinstance(curr, BackgroundSpec):
                        curr = curr.target

    def bg_step(self, *_):
        target = next(self.bg_sequence)
        bgcolor, callback, duration = None, None, None
        if isinstance(target, BackgroundSpec):
            target, bgcolor, callback, duration = target

        if callback:
            self.log.warn("BG Sequence received an item with a callback. "
                          "This is not supported and is ignored. {}".format(target))

        callback = self.bg_step
        spec = BackgroundSpec(target, bgcolor, callback, duration)
        self.log.debug("BG Sequence Step : {}".format(spec))
        self._bg_sequence_active = True
        self.bg = spec

    def background_sequence_set(self, targets):
        if not targets:
            targets = []

        _targets = []
        for target in targets:
            if not isinstance(target, BackgroundSpec):
                target = BackgroundSpec(target=target)
            provider = self._get_provider(target.target)
            if not provider:
                self.log.warn("Provider not found for background {}. Not Using.".format(target.target))
            else:
                _targets.append(target)

        self._bg_sequence_persistence.update(_targets)
        self.bg_update()

    def bg_update(self):
        sequence_targets = self._bg_sequence_persistence.get()
        if len(sequence_targets) > 1:
            self.bg_sequence = sequence_targets
        elif len(sequence_targets) == 1:
            bg = sequence_targets[0]
            if bg.duration or bg.callback:
                bg = BackgroundSpec(target=bg.target,
                                    bgcolor=bg.bgcolor,
                                    callback=None,
                                    duration=None)
            self._bg_sequence_active = False
            self.bg = bg
        else:
            self._bg_sequence_active = False
            super(BackgroundSequenceMixin, self).bg_update()





try:
    from .importer import BackgroundSequenceImporterMixin
    EffectiveBackgroundMixin = BackgroundSequenceImporterMixin
except ImportError:
    from .basic import BackgroundSequenceMixin
    EffectiveBackgroundMixin = BackgroundSequenceMixin





from .models import BackgroundModel
from .models import metadata

from ebs.linuxnode.db.sequence import GenericSequencePersistenceManager


class BackgroundSequencePersistenceManager(GenericSequencePersistenceManager):
    _db_name = 'background'
    _db_model = BackgroundModel
    _db_metadata = metadata

    def _insert_item(self, seq, item):
        # Item is to be inserted into the resource_manager by the caller.
        # The caller should also ensure the resources are downloaded and
        # available before it is set.
        pass

    def _clear_item(self, item):
        if self._actual.resource_manager.has(item.target):
            r = self._actual.resource_manager.get(item.target)
            r.rtype = None
            r.commit()





from twisted.internet import reactor
from ebs.linuxnode.gui.kivy.core.basenode import BaseIoTNodeGui
from ebs.linuxnode.gui.kivy.mediaplayer.mixin import MediaPlayerGuiMixin
from ebs.linuxnode.bgsequence.mixin import EffectiveBackgroundMixin
from kivy_garden.ebs.clocks.digital import SimpleDigitalClock
from ebs.linuxnode.core.background import BackgroundSpec

try:
    from ebs.linuxnode.gui.kivy.mediaplayer.omxplayer import OMXPlayerGuiMixin
    BaseNode = OMXPlayerGuiMixin
except ImportError:
    BaseNode = MediaPlayerGuiMixin


class ExampleNode(EffectiveBackgroundMixin, BaseNode, BaseIoTNodeGui):
    def _set_bg_sequence(self, targets):
        self.bg_sequence = targets

    @property
    def clock(self):
        return SimpleDigitalClock()

    _bgseries = [
        BackgroundSpec('1.0:0.5:0.5:1.0', duration=10),
        BackgroundSpec('image.jpg', duration=10),
        'video.mp4',
        'pdf.pdf',
        BackgroundSpec('structured:clock', duration=10),
        BackgroundSpec(None, duration=10),
    ]

    _bgseries_2 = [
        BackgroundSpec('1.0:0.5:0.5:1.0', duration=10),
        'video.mp4',
        BackgroundSpec('structured:clock', duration=10),
        BackgroundSpec(None, duration=10),
    ]

    def _background_sequence_example(self):
        reactor.callLater(5, self._set_bg_sequence, self._bgseries)

    def _background_sequence_set_example(self):
        reactor.callLater(5, self.background_sequence_set, self._bgseries)
        reactor.callLater(42, self.background_sequence_set, self._bgseries_2)

    def start(self):
        super(ExampleNode, self).start()
        self._background_sequence_set_example()
